import { DataType } from "../types";
export declare function buffer_to_base64(buffer: ArrayBuffer): string;
export declare function base64_to_buffer(base64: string): ArrayBuffer;
export declare function swap(buffer: ArrayBuffer, dtype: DataType): void;
//# sourceMappingURL=buffer.d.ts.map