export const root: string
export const tabs_header: string
export const btn_group: string
export const btn: string
export const headers_wrapper: string
export const above: string
export const right: string
export const below: string
export const left: string
export const headers: string
export const tab: string
export const active: string
export const close: string
export default ""
