export const root = "bk-root"
export const clearfix = "bk-clearfix"
export default `.bk-root .bk-clearfix:before,.bk-root .bk-clearfix:after{content:"";display:table;}.bk-root .bk-clearfix:after{clear:both;}`
